' Copyright 2008 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

''' <summary>
''' Class which stores a NMEA parser
''' </summary>
Public Class NMEAParser
    ''' <summary>
    ''' <c>true</c>if check for GPS sentence checksum is ignored / not calculated 
    ''' </summary>
    Public noChecksum As Boolean = False

    Private Shared usNumberFormat As New System.Globalization.CultureInfo("en-us")

    ''' <summary>
    ''' <c>false</c> indicates no RMC sentence received, so we are going to "fallback" to parse position from GGA sentence
    ''' <c>true</c> RMC sentence received, turns off position parsing in GGA sentence parser
    ''' </summary>
    Private rmcReceived As Boolean = False

    ''' <summary>
    ''' Parses the specified sentence and return the updated position in result.
    ''' </summary>
    ''' <param name="sentence">The sentence to parse.</param>
    ''' <param name="result">The resulting GPS-Position.</param>
    ''' <returns></returns>
    Public Function Parse(ByVal sentence As String, ByRef result As GPS_POSITION) As Boolean
        If IsValidSentence(sentence) = False Then
            Return False
        End If

        Select Case sentence.Substring(0, 6)
            Case "$GPRMC"
                Return ParseRMC(sentence, result)
            Case "$GPGGA"
                Return ParseGGA(sentence, result)
            Case "$GPGSA"
                Return ParseGSA(sentence, result)
            Case Else
                Return False
        End Select
    End Function
    ''' <summary>
    ''' Determines whether the specified sentence is a valid sentence.
    ''' </summary>
    ''' <param name="sentence">The sentence to check.</param>
    ''' <returns>
    ''' <c>true</c> if whether the specified sentence is a valid sentence; otherwise, <c>false</c>.
    ''' </returns>
    Private Function IsValidSentence(ByRef sentence As String) As Boolean
        If sentence.StartsWith("$GP") Then
            Dim asterisk As Integer = sentence.IndexOf("*")
            If asterisk <= 5 Then
                Return False
            End If
            If noChecksum Then
                sentence = sentence.Substring(0, asterisk - 1)
                Return True
            End If
            Dim checksum As Integer = Integer.Parse(sentence.Substring(asterisk + 1, 2), Globalization.NumberStyles.AllowHexSpecifier)
            sentence = sentence.Substring(0, asterisk)
            Dim calculatedChecksum As Integer = 0
            For i As Integer = 1 To asterisk - 1
                calculatedChecksum = calculatedChecksum Xor Asc(sentence.Chars(i))
            Next
            Return checksum = calculatedChecksum
        Else
            Return False
        End If
    End Function
    ''' <summary>
    ''' Parses a NMEA RMC sentence (The Recommended Minimum).
    ''' </summary>
    ''' <param name="rmcString">The RMC Sentence.</param>
    ''' <returns>always true</returns>
    Private Function ParseRMC(ByRef rmcString As String, ByRef result As GPS_POSITION) As Boolean

        ' set to true means no position parsing in GGA sentence
        rmcReceived = True

        Dim part As String() = SplitSentence(rmcstring)

        'extract time and date if available
        If Not part(1) = "" AndAlso Not part(9) = "" Then
            result.SetUTCTime(2000 + Integer.Parse(part(9).Substring(4, 2)), Integer.Parse(part(9).Substring(2, 2)), Integer.Parse(part(9).Substring(0, 2)), Integer.Parse(part(1).Substring(0, 2)), Integer.Parse(part(1).Substring(2, 2)), Integer.Parse(part(1).Substring(4, 2)))
        End If

        If Not part(3) = "" AndAlso Not part(4) = "" AndAlso Not part(5) = "" AndAlso Not part(6) = "" Then
            ParseGPSPosition(part(3), part(4), part(5), part(6), result)
        End If
        If Not part(7) = "" Then
            result.Speed = Single.Parse(part(7), usNumberFormat)
        End If
        If Not part(8) = "" Then
            result.Heading = Single.Parse(part(8), usNumberFormat)
        End If
        If part(2) = "A" Then
            ' valid = FIX
            result.FixType = 1
        Else
            ' invalid = NOFIX
            result.FixType = 0
        End If
        Return True
    End Function
    ''' <summary>
    ''' Parses a NMEA GGA sentence (essential fix data which provide 3D location and accuracy data).
    ''' </summary>
    ''' <param name="ggaString">The GGA sentence.</param>
    ''' <returns>always true</returns>
    Private Function ParseGGA(ByRef ggaString As String, ByRef result As GPS_POSITION) As Boolean
        Dim part As String() = SplitSentence(ggastring)

        If rmcReceived = False Then
            ' no RMC sentence received ever, so we are going to get position and GPS fix info from GGA sentence
            ' there is NO full time and date support, so we cannot adjust time based on GPS time
            ' also there is NO heading info, and SPEED over ground
            ' but at least POSITION and FIX we can get from GGA 

            ' part 2 to 5  same format as in RMC for position
            If Not part(2) = "" AndAlso Not part(3) = "" AndAlso Not part(4) = "" AndAlso Not part(5) = "" Then
                ParseGPSPosition(part(2), part(3), part(4), part(5), result)
            End If

            'info from http://campar.in.tum.de/twiki/pub/Chair/NaviGpsDemon/nmea.html#GGA
            'Fix quality:  0 = invalid
            '              1 = GPS fix (SPS)
            '              2 = DGPS fix
            '              3 = PPS fix
            '              4 = Real Time Kinematic
            '              5 = Float RTK
            '              6 = estimated (dead reckoning) (2.3 feature)
            '              7 = Manual input mode
            '              8 = Simulation mode
            result.FixType = Integer.Parse(part(6))
            ParseGGA = True
        Else
            ParseGGA = False
        End If

        If Not part(7) = "" Then
            result.SatelliteCount = Integer.Parse(part(7))
        End If

        If Not part(9) = "" AndAlso part(10) = "M" Then
            result.AltitudeWRTSeaLevel = CInt(Double.Parse(part(9), usNumberFormat))
        End If
    End Function
    ''' <summary>
    ''' Parses a NMEA GSA sentence (GPS DOP and active satellites).
    ''' </summary>
    ''' <param name="gsaString">The GSA sentence.</param>
    ''' <returns>always false</returns>
    Private Function ParseGSA(ByRef gsaString As String, ByRef result As GPS_POSITION) As Boolean
        Dim part As String() = SplitSentence(gsastring)

        If Not part(15) = "" Then
            result.PositionDilutionOfPrecision = Single.Parse(part(15), usNumberFormat)
        End If
        If Not part(16) = "" Then
            result.HorizontalDilutionOfPrecision = Single.Parse(part(16), usNumberFormat)
        End If
        If Not part(17) = "" Then
            result.VerticalDilutionOfPrecision = Single.Parse(part(17), usNumberFormat)
        End If
        Return False
    End Function
    ''' <summary>
    ''' Parses a GPS position and updates the GPS_POSITION
    ''' </summary>
    ''' <param name="latitudeString">NMEA latitude string</param>
    ''' <param name="northSouth">_N_orth or _S_outh</param>
    ''' <param name="longitudeString">NMEA longitude string</param>
    ''' <param name="eastWest">_E_ast or _W_est</param>
    ''' <param name="result">The resulting GPS_POSITION.</param>
    Private Sub ParseGPSPosition(ByVal latitudeString As String, ByVal northSouth As String, ByVal longitudeString As String, ByVal eastWest As String, ByRef result As GPS_POSITION)
        ' get axis from degree if possible
        Dim degrees As Double = Double.Parse(latitudeString, usNumberFormat)
        Dim hours As Double = Math.Floor(degrees / 100)
        Dim minutes As Double = degrees Mod 100
        result.Latitude = hours + minutes / 60
        If northSouth = "S" Then result.Latitude *= -1

        degrees = Double.Parse(longitudeString, usNumberFormat)
        hours = Math.Floor(degrees / 100)
        minutes = degrees Mod 100
        result.Longitude = hours + minutes / 60
        If eastWest = "W" Then result.Longitude *= -1
    End Sub
    ''' <summary>
    ''' Splits a sentence (sentence contain comma separated columns).
    ''' </summary>
    ''' <param name="sentence">The sentence.</param>
    ''' <returns>An array of the columns.</returns>
    Public Function SplitSentence(ByRef sentence As String) As String()
        Return sentence.Split(","c)
    End Function
End Class
