' Copyright 2008 Jozef Vince <jozef.vince@freemap.sk>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

#If Not Win32 Then
Imports WaveAudio
#End If
Imports System.IO

''' <summary>
''' Encapulate various sound player methods
''' </summary>
Public Class SoundPlayer
#If Not Win32 Then
    ''' <summary>
    ''' Instance of WaveOut for playing sounds
    ''' </summary>
    Private mainSound As WaveOut = New WaveOut
#Else
    Private mainSound As New Microsoft.VisualBasic.Devices.Audio
#End If

    ''' <summary>
    ''' If <b>true</b> sounds are enabled otherwise operate silently
    ''' </summary>
    Private soundEnabled As Boolean

    ''' <summary>
    ''' While False other request to play sounds are ignored
    ''' </summary>
    Private isLoud As Boolean = True

    ''' <summary>
    ''' Mute request counter - track number of mute/unmute calls
    ''' </summary>
    Private muteCounter As Integer = 0

    ''' <summary>
    ''' Gets or sets a value indicating whether sound output are enabled.
    ''' </summary>
    ''' <value><c>true</c> if [enable sounds]; otherwise, <c>false</c>.</value>
    Public Property IsEnabled() As Boolean
        Get
            IsEnabled = soundEnabled
        End Get
        Set(ByVal value As Boolean)
            soundEnabled = value
            If soundEnabled = False Then
                isLoud = False
            Else
                If muteCounter < 1 Then
                    isLoud = True
                End If
            End If
        End Set
    End Property


    ''' <summary>
    ''' Mutes sounds.
    ''' </summary>
    Private Sub Mute()
        muteCounter += 1
        If muteCounter > 0 Or soundEnabled = False Then
            isLoud = False
        End If
    End Sub

    ''' <summary>
    ''' Unmute the sounds.
    ''' </summary>
    Private Sub UnMute()
        muteCounter -= 1
        If muteCounter < 1 And soundEnabled = True Then
            isLoud = True
        End If
    End Sub

    ''' <summary>
    ''' Play "fix aquired" sound.
    ''' </summary>
    Public Sub FixAquired()
        If isLoud Then
            SafePlay(GetResourceStream("fixaquired.wav"), 0, &HFFFF, &HFFFF)
        End If
    End Sub

    ''' <summary>
    ''' Play "fix lost" sound.
    ''' </summary>
    Public Sub FixLost()
        If isLoud Then
            SafePlay(GetResourceStream("fixlost.wav"), 0, &HFFFF, &HFFFF)
        End If
    End Sub

    ''' <summary>
    ''' Play click sound.
    ''' </summary>
    Public Sub Click()
        If isLoud Then
            ' http://www.pdsounds.org/sounds/blip
            ' http://www.pdsounds.org/audio/download/209/bip.mp3
            SafePlay(GetResourceStream("bip.wav"), 0, &HFFFF, &HFFFF)
        End If
    End Sub

    ''' <summary>
    ''' Play blip sound.
    ''' </summary>
    Public Sub Blip()
        If isLoud Then
            SafePlay(GetResourceStream("blip1.wav"), 0, &HFFFF, &HFFFF)
        End If
    End Sub

    ''' <summary>
    ''' Blips at start of recording and mute all sounds.
    ''' </summary>
    Public Sub BlipRecordStart()
        ' when going to record voice disable sound before blip
        Mute()

        If soundEnabled Then
            SafePlay(GetResourceStream("blip1.wav"), 0, &HFFFF, &HFFFF)
        End If
    End Sub

    ''' <summary>
    ''' Blip at end of recording, and unmute.
    ''' </summary>
    Public Sub BlipRecordStop()
        If soundEnabled Then
            SafePlay(GetResourceStream("blip1.wav"), 0, &HFFFF, &HFFFF)
        End If

        ' When record stopped enable sound after blip.
        UnMute()
    End Sub

    ''' <summary>
    ''' Plays the SoundFile or click if no SoundFile specified.
    ''' </summary>
    ''' <param name="SoundFile">The sound file.</param>
    Public Sub PlayClick(Optional ByVal SoundFile As String = "")
        If SoundFile = "" Then
            Click()
        Else
            PlayClip(SoundFile, False)
        End If
    End Sub

    ''' <summary>
    ''' Plays the SoundFile or blip if no SoundFile specified.
    ''' </summary>
    ''' <param name="SoundFile">The sound file.</param>
    Public Sub PlayBlip(Optional ByVal SoundFile As String = "")
        If SoundFile = "" Then
            Blip()
        Else
            PlayClip(SoundFile, False)
        End If
    End Sub


    ''' <summary>
    ''' Plays the clip.
    ''' </summary>
    ''' <param name="fileName">Name of the file.</param>
    ''' <param name="waitFor">if set to <c>true</c> simulates synchronous play mode.</param>
    Public Sub PlayClip(ByVal fileName As String, Optional ByVal waitFor As Boolean = True)
        If isLoud Then
            Mute()

            If File.Exists(fileName) Then
#If Not Win32 Then
                Dim Player As WaveOut = New WaveOut
                Try
                    Player.Play(fileName, 0, &HFFFF, &HFFFF)
                    If waitFor Then
                        Dim milisecondsWait As Integer
                        milisecondsWait = Player.Milliseconds()
                        System.Threading.Thread.Sleep(milisecondsWait + 100)
                    End If
                Catch ex As Exception
                    MsgBox("Unable to play voice clip:" & ex.Message)
                End Try
#Else
                If waitFor Then
                    mainSound.Play(fileName, AudioPlayMode.WaitToComplete)
                Else
                    mainSound.Play(fileName, AudioPlayMode.Background)
                End If
#End If
            End If
            UnMute()
        End If
    End Sub
#If Not Win32 Then
    ''' <summary>
    ''' Wraper for exception aware Play.
    ''' </summary>
    ''' <param name="sourceStream">The source stream.</param>
    ''' <param name="bufferSize">Size of the buffer.</param>
    ''' <param name="volLeft">Left channel volume.</param>
    ''' <param name="volRight">Right channel volue.</param>
    ''' <returns></returns>
    Private Function SafePlay(ByRef sourceStream As Stream, ByVal bufferSize As Integer, ByVal volLeft As Integer, ByVal volRight As Integer) As Wave.MMSYSERR
        Try
            mainSound.Play(sourceStream, bufferSize, volLeft, volRight)
        Catch ex As Exception
            MsgBox("Unable to play: " & ex.Message)
        End Try
    End Function
#Else
    Private Function SafePlay(ByRef sourceStream As Stream, ByVal bufferSize As Integer, ByVal volLeft As Integer, ByVal volRight As Integer) As Integer
        Try
            mainSound.Play(sourceStream, AudioPlayMode.Background)
        Catch ex As Exception
            MsgBox("Unable to play: " & ex.Message)
        End Try
    End Function
#End If
End Class
