' Copyright 2008 - 2009 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

' REMEMBER: as soon as addTrack or addTrackPoint get called from different threads, they must have a mutex!

Option Explicit On
Option Strict On

Imports System.IO
Imports System.Collections.Generic

''' <summary>
''' Class for general handling of a GPX file.
''' </summary>
Public Class LoggingGPXFile
    Inherits GPXFile

    Private currentTrackPoints As SingleLinkedList(Of OSMTrackPoint) = New SingleLinkedList(Of OSMTrackPoint)
    Private wayPoints As SingleLinkedList(Of OSMWayPoint) = New SingleLinkedList(Of OSMWayPoint)
    Private firstTrack As Boolean = True
    Private filename As String
    Private gpxWithDoP As Boolean = False
    Private wayPointStreamWriter As StreamWriter = Nothing
    Private trackPointStreamWriter As StreamWriter = Nothing
    Public Sub New(ByVal filename As String, ByVal storeDoP As Boolean)
        Me.filename = filename
        Me.gpxWithDoP = storeDoP
    End Sub
    ''' <summary>
    ''' Creates an empty WayPoint file
    ''' </summary>
    Private Sub CreateWayPointFile()
        If wayPointStreamWriter IsNot Nothing Then
            wayPointStreamWriter.Close()
        End If
        wayPointStreamWriter = New StreamWriter(New FileStream(filename & "-wp." & GPX_EXTENSION, FileMode.Create))
        WriteGPXHeader(wayPointStreamWriter)
    End Sub
    ''' <summary>
    ''' Creates an empty TrackPoint file
    ''' </summary>
    Private Sub CreateTrackPointFile()
        trackPointStreamWriter = New StreamWriter(New FileStream(filename & "-tp." & GPX_EXTENSION, FileMode.Create))
        WriteGPXHeader(trackPointStreamWriter)
    End Sub
    ''' <summary>
    ''' Adds and starts a new track in the GPX file.
    ''' </summary>
    Public Overrides Sub AddTrack()
        If trackPointStreamWriter Is Nothing OrElse (firstTrack = False AndAlso currentTrackPoints.Count = 0) Then Exit Sub
        If firstTrack = False Then
            trackPointStreamWriter.WriteLine("</trkseg></trk>")
            ' make sure at least the last track is complete
            trackPointStreamWriter.Flush()
            currentTrackPoints = New SingleLinkedList(Of OSMTrackPoint)
        End If
        firstTrack = False
        trackPointStreamWriter.WriteLine("<trk><trkseg>")
    End Sub
    ''' <summary>
    ''' Adds a new waypoint (Point Of Interest) to the GPX file.
    ''' </summary>
    ''' <param name="position">The position.</param>
    ''' <param name="time">The time.</param>
    ''' <param name="name">The name.</param>
    ''' <param name="comment">The comment.</param>
    ''' <param name="linkhref">The linkhref.</param>
    Public Overrides Sub AddWayPoint(ByRef position As GPS_POSITION, ByVal time As Date, ByVal name As String, ByVal comment As String, ByVal linkhref As String)
        waypoint_mutex.WaitOne()
        If wayPointStreamWriter Is Nothing Then
            CreateWayPointFile()
            wayPointStreamWriter.AutoFlush = True
        End If
        wayPoints.Add(New OSMWayPoint(position, time, name, comment, linkhref))
        wayPointStreamWriter.Write(wayPoints.GetLast.ToXML(gpxWithDoP))
        waypoint_mutex.ReleaseMutex()
    End Sub
    ''' <summary>
    ''' Adds a trackpoint to the current track of the GPX file.
    ''' </summary>
    ''' <param name="position">The position.</param>
    ''' <param name="time">The time.</param>
    Public Overrides Sub AddTrackPoint(ByVal position As GPS_POSITION, ByVal time As Date)
        If trackPointStreamWriter Is Nothing Then
            CreateTrackPointFile()
        End If
        If firstTrack Then
            AddTrack()
        End If
        currentTrackPoints.Add(New OSMTrackPoint(New MAP_POSITION(position)))
        trackPointStreamWriter.Write("<trkpt lat=""" & position.Latitude.ToString("0.0000000", myCultureInfo) & """ lon=""" & position.Longitude.ToString("0.0000000", myCultureInfo) & """><ele>" & position.AltitudeWRTSeaLevel & "</ele><time>" & time.ToUniversalTime.ToString("yyyy-MM-ddTHH:mm:ssZ", myCultureInfo) & "</time>")
        If gpxWithDoP Then
            trackPointStreamWriter.Write("<hdop>" & position.HorizontalDilutionOfPrecision.ToString(myCultureInfo) & "</hdop><vdop>" & position.VerticalDilutionOfPrecision.ToString(myCultureInfo) & "</vdop><pdop>" & position.PositionDilutionOfPrecision.ToString(myCultureInfo) & "</pdop>")
        End If
        trackPointStreamWriter.WriteLine("</trkpt>")
        ' flush the trackPointBuffer every 50 trackpoints
        If GetTrackPointsCount() Mod 50 = 0 Then
            trackPointStreamWriter.Flush()
        End If
    End Sub
    ''' <summary>
    ''' Writes the GPX-XML header to a stream.
    ''' </summary>
    ''' <param name="stream">The stream.</param>
    Protected Sub WriteGPXHeader(ByVal stream As StreamWriter)
        stream.WriteLine("<?xml version=""1.0"" standalone=""yes""?>")
        stream.WriteLine("<gpx version=""1.1"" creator=""OSMtracker"" xmlns:xsi=""http://www.w3.org/2001/XMLSchema-instance"" xmlns=""http://www.topografix.com/GPX/1/1"" xsi:schemaLocation=""http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd"">")
    End Sub
    ''' <summary>
    ''' Saves the opened GPX file and closes it.
    ''' </summary>
    Public Overrides Sub Save()
        If wayPointStreamWriter IsNot Nothing Then
            wayPointStreamWriter.WriteLine("</gpx>")
            wayPointStreamWriter.Close()
        End If
        If trackPointStreamWriter IsNot Nothing Then
            If firstTrack = False Then
                trackPointStreamWriter.WriteLine("</trkseg></trk>")
            End If
            trackPointStreamWriter.WriteLine("</gpx>")
            trackPointStreamWriter.Close()
        End If
    End Sub
    ''' <summary>
    ''' Returns the number of waypoints.
    ''' </summary>
    ''' <returns></returns>
    Public Overrides Function GetWayPointsCount() As Integer
        Return wayPoints.Count
    End Function
    ''' <summary>
    ''' Returns the number track points of the current track.
    ''' </summary>
    ''' <returns></returns>
    Public Overrides Function GetTrackPointsCount() As Integer
        Return currentTrackPoints.Count
    End Function
    ''' <summary>
    ''' Removes the waypoint at position index.
    ''' </summary>
    ''' <param name="index">The index.</param>
    Public Overrides Function RemoveWayPoint(ByVal index As Integer) As Boolean
        Dim removed As Boolean = False
        waypoint_mutex.WaitOne()
        If wayPoints.Count > 0 Then
            If wayPoints.Remove(index) Then
                removed = True
                CreateWayPointFile()
                For Each wayPoint As OSMWayPoint In wayPoints
                    wayPointStreamWriter.Write(wayPoint.ToXML(gpxWithDoP))
                Next
                wayPointStreamWriter.Flush()
                wayPointStreamWriter.AutoFlush = True
            End If
        End If
        waypoint_mutex.ReleaseMutex()
        Return removed
    End Function
    ''' <summary>
    ''' Returns the waypoints as a LinkedList.
    ''' </summary>
    ''' <returns></returns>
    Public Overrides Function GetWayPoints() As SingleLinkedList(Of OSMWayPoint)
        Return wayPoints
    End Function
    ''' <summary>
    ''' Returns the trackpoints of the current track as a LinkedList.
    ''' </summary>
    ''' <returns></returns>
    Public Overrides Function GetTrackPoints() As SingleLinkedList(Of OSMTrackPoint)
        If firstTrack = False Then
            Return currentTrackPoints
        Else
            Return Nothing
        End If
    End Function
End Class
