' Copyright 2009 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

''' <summary>
''' Implementation for a SingleLinkedList
''' </summary>
Public Class SingleLinkedList(Of T)
    Implements ICollection(Of T), ICollection

    Private nodeCount As Integer
    Private listHead As SingleLinkedListNode(Of T)
    Private listTail As SingleLinkedListNode(Of T)

    ''' <summary>
    ''' Returns the item at the specified index.
    ''' </summary>
    ''' <param name="index">The index.</param>
    ''' <returns>The item stored in the list</returns>
    Public Function GetItem(ByVal index As Integer) As T
        If index < 0 Then
            Throw New ArgumentOutOfRangeException
        ElseIf index >= Count Then
            Throw New ArgumentOutOfRangeException
        End If
        Dim cnt As Integer = 0
        Dim tempNode As SingleLinkedListNode(Of T) = listHead
        While tempNode IsNot Nothing
            If cnt = index Then
                Return tempNode.value
            End If
            cnt = cnt + 1
            tempNode = tempNode.nextNode
        End While
        Return Nothing
    End Function
    ''' <summary>
    ''' Returns the last item stored in the list.
    ''' </summary>
    ''' <returns>the item</returns>
    Public Function GetLast() As T
        If listTail Is Nothing Then
            Return Nothing
        Else
            Return listTail.value
        End If
    End Function
    ''' <summary>
    ''' Removes the item stored at the specified index in the list.
    ''' </summary>
    ''' <param name="index">The index.</param>
    ''' <returns></returns>
    Public Function Remove(ByVal index As Integer) As Boolean
        If index < 0 Then
            Throw New ArgumentOutOfRangeException
        ElseIf index >= Count Then
            Throw New ArgumentOutOfRangeException
        End If
        Dim tempNode As SingleLinkedListNode(Of T) = listHead
        If index = 0 Then
            Return DeleteFirst()
        Else
            Dim cnt As Integer = 1
            While tempNode IsNot Nothing
                If cnt = index Then
                    Return DeleteNextNode(tempNode)
                End If
                cnt = cnt + 1
                tempNode = tempNode.nextNode
            End While
        End If
        Return False
    End Function
    ''' <summary>
    ''' Removes the specified item from the list.
    ''' </summary>
    ''' <param name="item">The item.</param>
    ''' <returns>True if the item was found and removed, false otherwise.</returns>
    Public Function Remove(ByVal item As T) As Boolean Implements ICollection(Of T).Remove
        If listHead IsNot Nothing AndAlso listHead.value.Equals(item) Then
            Return DeleteFirst()
        End If
        Dim tempNode As SingleLinkedListNode(Of T) = listHead
        While tempNode IsNot Nothing
            If tempNode.nextNode IsNot Nothing AndAlso tempNode.nextNode.value.Equals(item) Then
                Return DeleteNextNode(tempNode)
            End If
            tempNode = tempNode.nextNode
        End While
        Return False
    End Function
    ''' <summary>
    ''' Removes the last item on the list.
    ''' </summary>
    ''' <returns>False if the list was empty.</returns>
    Public Function RemoveLast() As Boolean
        Return Remove(Count - 1)
    End Function
    ''' <summary>
    ''' Deletes the first item in the list.
    ''' </summary>
    ''' <returns>False if the list was empty.</returns>
    Public Function DeleteFirst() As Boolean
        If listHead IsNot Nothing Then
            If listHead.nextNode Is Nothing Then
                listHead = Nothing
                listTail = Nothing
                nodeCount = 0
            Else
                listHead = listHead.nextNode
                nodeCount = nodeCount - 1
            End If
            Return True
        Else
            Return False
        End If
    End Function
    ''' <summary>
    ''' Deletes the next node.
    ''' </summary>
    ''' <param name="node">The node.</param>
    ''' <returns>true if the removal was successfull</returns>
    Private Function DeleteNextNode(ByRef node As SingleLinkedListNode(Of T)) As Boolean
        If node IsNot Nothing AndAlso node.nextNode IsNot Nothing Then
            If ReferenceEquals(node.nextNode, listTail) Then
                listTail = node
            End If
            node.nextNode = node.nextNode.nextNode
            nodeCount = nodeCount - 1
            Return True
        Else
            Return False
        End If
    End Function
    ''' <summary>
    ''' Determines whether a value is in the LinkedList. 
    ''' </summary>
    ''' <param name="item">he value to locate in the SingleLinkedList. The value can be a null reference (Nothing in Visual Basic) for reference types.</param>
    ''' <returns>
    ''' true if value is found in the LinkedList; otherwise, false. 
    ''' </returns>
    Public Function Contains(ByVal item As T) As Boolean Implements ICollection(Of T).Contains
        Dim tempNode As SingleLinkedListNode(Of T) = listHead
        While tempNode IsNot Nothing
            If tempNode.value.Equals(item) Then
                Return True
            End If
            tempNode = tempNode.nextNode
        End While
        Return False
    End Function
    ''' <summary>
    ''' Copies the entire SingleLinkedList to a compatible one-dimensional <see cref="T:System.Array"></see>, starting at the specified index of the target array. 
    ''' </summary>
    ''' <param name="array">The one-dimensional <see cref="T:System.Array"></see> that is the destination of the elements copied from SingleLinkedList. The Array must have zero-based indexing.</param>
    ''' <param name="arrayIndex">The zero-based index in array at which copying begins.</param>
    ''' <exception cref="T:System.ArgumentNullException">array is a null reference (Nothing in Visual Basic).</exception>
    ''' <exception cref="T:System.ArgumentException">index is equal to or greater than the length of array -or The number of elements in the source SingleLinkedList is greater than the available space from index to the end of the destination array.</exception>
    ''' <exception cref="T:System.ArgumentOutOfRangeException">index is less than zero.</exception>
    Public Sub CopyTo(ByVal array As Array, ByVal arrayIndex As Integer) Implements ICollection.CopyTo
        If array Is Nothing Then
            Throw New ArgumentNullException
        ElseIf arrayIndex < 0 Then
            Throw New ArgumentOutOfRangeException
        ElseIf arrayIndex >= array.Length OrElse array.Length < arrayIndex + Count Then
            Throw New ArgumentException
        Else
            Dim tempNode As SingleLinkedListNode(Of T) = listHead
            While tempNode IsNot Nothing
                'array(arrayIndex) = tempNode.item
                arrayIndex = arrayIndex + 1
                tempNode = tempNode.nextNode
            End While
        End If
    End Sub
    ''' <summary>
    ''' Copies the entire SingleLinkedList to a compatible one-dimensional <see cref="T:System.Array"></see>, starting at the specified index of the target array. 
    ''' </summary>
    ''' <param name="array">The one-dimensional <see cref="T:System.Array"></see> that is the destination of the elements copied from SingleLinkedList. The Array must have zero-based indexing.</param>
    ''' <param name="arrayIndex">The zero-based index in array at which copying begins.</param>
    ''' <exception cref="T:System.ArgumentNullException">array is a null reference (Nothing in Visual Basic).</exception>
    ''' <exception cref="T:System.ArgumentException">index is equal to or greater than the length of array -or The number of elements in the source SingleLinkedList is greater than the available space from index to the end of the destination array.</exception>
    ''' <exception cref="T:System.ArgumentOutOfRangeException">index is less than zero.</exception>
    Public Sub CopyTo(ByVal array As T(), ByVal arrayIndex As Integer) Implements ICollection(Of T).CopyTo
        If array Is Nothing Then
            Throw New ArgumentNullException
        ElseIf arrayIndex < 0 Then
            Throw New ArgumentOutOfRangeException
        ElseIf arrayIndex >= array.Length OrElse array.Length < arrayIndex + Count Then
            Throw New ArgumentException
        Else
            Dim tempNode As SingleLinkedListNode(Of T) = listHead
            While tempNode IsNot Nothing
                array(arrayIndex) = tempNode.value
                arrayIndex = arrayIndex + 1
                tempNode = tempNode.nextNode
            End While
        End If
    End Sub
    ''' <summary>
    ''' Copies the entire SingleLinkedList to a one-dimensional <see cref="T:System.Array"></see>.
    ''' </summary>
    ''' <returns>The array with the values from the SingleLinkeList.</returns>
    Public Function ToArray() As T()
        Dim array(Count - 1) As T
        CopyTo(array, 0)
        Return array
    End Function
    ''' <summary>
    ''' Removes all nodes from the SingleLinkedList.
    ''' </summary>
    Public Sub Clear() Implements ICollection(Of T).Clear
        nodeCount = 0
        listHead = Nothing
        listTail = Nothing
    End Sub
    ''' <summary>
    ''' Adds a new value at the end of the SingleLinkedList.
    ''' </summary>
    ''' <param name="item">The value to add at the end of the SingleLinkedList.</param>
    Public Sub Add(ByVal item As T) Implements ICollection(Of T).Add
        nodeCount = nodeCount + 1
        If listTail Is Nothing Then
            listTail = New SingleLinkedListNode(Of T)
            listTail.value = item
            listHead = listTail
        Else
            listTail.nextNode = New SingleLinkedListNode(Of T)
            listTail.nextNode.value = item
            listTail = listTail.nextNode
        End If
    End Sub
    ''' <summary>
    ''' Gets the number of nodes actually contained in the SingleLinkedList.
    ''' </summary>
    ''' <value></value>
    ''' <returns>The number of nodes actually contained in the LinkedList.</returns>
    Public ReadOnly Property Count() As Integer Implements ICollection.Count, ICollection(Of T).Count
        Get
            Return nodeCount
        End Get
    End Property
    ''' <summary>
    ''' Returns an enumerator that iterates through the SingleLinkedList.
    ''' </summary>
    ''' <returns>
    ''' An SingleLinkedListEnumerator for the SingleLinkedList.
    ''' </returns>
    Public Function GetEnumerator() As IEnumerator(Of T) Implements IEnumerable(Of T).GetEnumerator
        Return New SingleListEnumerator(Of T)(listHead)
    End Function
    ''' <summary>
    ''' Returns an enumerator that iterates through the SingleLinkedList.
    ''' </summary>
    ''' <returns>
    ''' An SingleLinkedListEnumerator for the SingleLinkedList.
    ''' </returns>
    Public Function GetEnumerator1() As IEnumerator Implements IEnumerable.GetEnumerator
        Return Nothing
    End Function
    ''' <summary>
    ''' Gets a value indicating whether the <see cref="T:System.Collections.ICollection"></see> is read-only.
    ''' </summary>
    ''' <value></value>
    ''' <returns>true if the <see cref="T:System.Collections.Generic.ICollection`1"></see> is read-only; otherwise, false. In the default implementation of SingleLinkedList, this property always returns false.</returns>
    Public ReadOnly Property IsReadOnly() As Boolean Implements ICollection(Of T).IsReadOnly
        Get
            Return False
        End Get
    End Property
    ''' <summary>
    ''' Gets a value indicating whether access to the <see cref="T:System.Collections.ICollection"></see> is synchronized (thread safe).
    ''' </summary>
    ''' <value></value>
    ''' <returns>true if access to the <see cref="T:System.Collections.Generic.ICollection`1"></see> is synchronized (thread safe); otherwise, false. In the default implementation of SingleLinkedList, this property always returns false.</returns>
    Public ReadOnly Property IsSynchronized() As Boolean Implements ICollection.IsSynchronized
        Get
            Return False
        End Get
    End Property
    ''' <summary>
    ''' Gets an object that can be used to synchronize access to the <see cref="T:System.Collections.ICollection"></see>.
    ''' </summary>
    ''' <value></value>
    ''' <returns>An object that can be used to synchronize access to the <see cref="T:System.Collections.ICollection"></see>. In the default implementation of SingleLinkedList, this property always returns the current instance.</returns>
    Public ReadOnly Property SyncRoot() As Object Implements ICollection.SyncRoot
        Get
            Return Me
        End Get
    End Property
End Class
